from enum import Enum
from typing import Optional

from pydantic import BaseModel, field_validator


class MeasureFunctionType(str, Enum):
    count = "count"
    countDistinct = "countDistinct"
    sum = "sum"
    average = "average"
    aggregate = "aggregate"
    max = "max"


class DimensionFunctionType(str, Enum):
    year = "year"
    month = "month"
    day = "day"
    yearMonth = "yearMonth"
    distinct = "distinct"


class WildCardOptions(str, Enum):
    contains = "contains"
    starts_with = "starts_with"
    ends_with = "ends_with"
    exactly_matches = "exactly_matches"
    formula = "formula"
    equals = "equals"
    not_equals = "not_equals"
    greater_than = "greater_than"
    less_than = "less_than"
    greater_than_equal_to = "greater_than_equal_to"
    less_than_equal_to = "less_than_equal_to"


class RangeOptions(str, Enum):
    none = None
    range = "range"
    at_least = "at_least"
    at_most = "at_most"


class FieldType(str, Enum):
    source = "source"
    calculated_field = "calculated_field"


class CategoryType(str, Enum):
    dimension = "dimension"
    measure = "measure"


class SortDirection(str, Enum):
    asc = "asc"
    desc = "desc"


class JoinOpt(str, Enum):
    all = "all"
    common = "common"


class SelectionModel(BaseModel):
    id: int | str
    exclude: Optional[bool] = None
    null: Optional[bool] = None
    values: Optional[list[str | int | float]] = None
    dimFunc: Optional[DimensionFunctionType] = None


class WildCardModel(BaseModel):
    id: int | str
    value: str
    exclude: Optional[bool] = None
    option: WildCardOptions = WildCardOptions.contains
    dimFunc: Optional[DimensionFunctionType] = None


class ConditionModel(BaseModel): ...


class SortByModel(BaseModel):
    id: int | str
    alias: Optional[str] = None
    direction: SortDirection = SortDirection.asc


class FilterModel(BaseModel):
    selections: Optional[list[SelectionModel]] = []
    wildcards: Optional[list[WildCardModel]] = []
    conditions: Optional[list[ConditionModel]] = []
    limit: Optional[int] = None
    offset: Optional[int] = None
    sort_by: Optional[list[SortByModel]] = []


class FieldsModel(BaseModel):
    id: Optional[int | str] = None
    name: str
    type: FieldType = FieldType.source
    category: CategoryType = CategoryType.dimension
    measure_func: MeasureFunctionType = MeasureFunctionType.count
    dimension_func: Optional[DimensionFunctionType] = None
    is_pii: bool = False
    sql_code: Optional[str] = None
    join_opt: Optional[JoinOpt] = None


class RangeModel(BaseModel):
    id: int | str
    alias: str
    value: tuple[int, int]
    option: RangeOptions


class MeasureFilterModel(BaseModel):
    range: list[RangeModel] = []


class ETLModel(BaseModel):
    name: str
    fields: list[FieldsModel]
    filter: Optional[FilterModel] = None
    measure_filter: Optional[MeasureFilterModel] = None
    cart_id: int = 0
    join: Optional[dict] = None


class FieldDetailsModel(BaseModel):
    id: int | str
    name: str
    type: FieldType = FieldType.source
    datatype_l1: str
    datatype_l2: str
    sql_code: str
    is_pii: bool
    asset_id: int | str
    asset_name: str
    asset_details: dict
    connection_id: int | str
    connection_source_name: str
    connection_credentials: dict
    count: Optional[int] = 1
    distinct_count: Optional[int] = 1


class SinkModel(BaseModel):
    connection_id: int
    connection_source_name: str
    connection_credentials: dict
    destination: str


class LinkModel(BaseModel):
    id: int | str
    source_field_id: int | str
    target_field_id: int | str
    source_asset_id: int | str
    target_asset_id: int | str
    source_count: int = 1
    target_count: int = 1
    source_count_distinct: int = 1
    target_count_distinct: int = 1
    records_mapped: int = 1
    score: Optional[int] = None
    ignore: Optional[bool] = False

    @field_validator("records_mapped", mode="before")
    def default_if_none(cls, v):
        return v if v is not None else 1



import logging
import os
import pickle

from datetime import datetime

import numpy as np
import pandas as pd

from intugle.core.settings import settings

log = logging.getLogger(__name__)


class L1Model:
    
    model_objects_directory = os.path.join(settings.MODEL_DIR_PATH, "model_objects", "datatype_l1_identification")
    model_results_directory = os.path.join(settings.MODEL_RESULTS_PATH, 'datatype_l1_identification')

    def __init__(self, *args, **kwargs):
        if not os.path.exists(self.model_objects_directory):
            raise FileNotFoundError(f"No models at {self.model_objects_directory}")
        
    # ====================================================================
    # group_labels: 
    #  - groups the similar datatypes into one single type
    # Parameters: 
    #     datatype- input datatype 
    # ====================================================================     
    def __group_labels(self, datatype: str) -> str:
        
        # groups the similar datatypes into one single type
        if datatype in ['Short_Integer', 'Long_Integer']:
            return 'Integer'
        elif datatype in ['Short_Float', 'Long_Float']:
            return 'Float'
        elif datatype in ['Short_Alphanumeric', 'Long_Alphanumeric']:
            return 'Alphanumeric'
        elif datatype in ['Open_ended_long_text', 'Open_ended_short_text']:
            return 'Open_ended_text'
        elif datatype == 'Close_ended_short_text':
            return 'Close_ended_text'
        else:
            return datatype
        
    # ====================================================================
    # feature_subset: 
    #  - Returns the list of required features for the modeling
    # ====================================================================             
    def feature_subset(self) -> list:
        
        log.info('Subsetting the char, par, word, rest features')
        feats_cols = ['dateRatio', 'wordlen_mean', 'rangeRatio', 'floatRatio', 'zero_flag', 'intRatio',
                      'alphaNumRatio', 'alphaRatio', 'frac_unique_sample', 'flag_numcells']
        
        # Returns the list of required features for the modeling
        return feats_cols
     
    # ====================================================================
    # model_prediction: 
    #  - Features subset from the dataset for prediction 
    #  - Loading the model objects and encoder objects for model prediction
    #  - Model prediction on the test dataset and inverse transform to get the labels for the encoder
    #  - Subset the required columns based on the process type and return the predicted dataframe
    # Parameters: 
    #    test_df - test dataframe with features for model prediction
    #    model_version - model version
    #    process_type - train/inference for getting the prediction    
    # ====================================================================                      
    def model_prediction(self, test_df: pd.DataFrame, model_version=datetime.now().strftime('%d%m%Y'),
                         process_type: str = 'train') -> pd.DataFrame:
                
        X_test = test_df.copy()
        
        del test_df
        
        # Features subset from the dataset for prediction          
        features_list = self.feature_subset()
            
        # Loading the model objects for model prediction
        # logger.debug('Reading the model objects')

# 
        model_path = os.path.join(self.model_objects_directory, 'di_l1_classifier_xgb_' + model_version + '.pkl')
        encoder_path = os.path.join(self.model_objects_directory, 'di_l1_classifier_encoder_' + model_version + '.pkl')
        with open(model_path, 'rb') as f:
            clf_model = pickle.load(f) 
        log.info("[*] Loaded L1 classifier model")
        # Loading the encoder objects for model prediction
        with open(encoder_path, 'rb') as f:
            encoder = pickle.load(f)
        log.info("[*] Loaded L1 Encoder")
        
        # Model prediction on the test dataset and inverse transform to get the labels for the encoder
        log.info('[*] Started the model prediction')
        
        predicted_model_prob = clf_model.predict_proba(X_test[features_list].values)
        predicted_prob_class = [(pred[np.argmax(pred)], np.argmax(pred)) for pred in predicted_model_prob]
        predicted_prob = [val[0] for val in predicted_prob_class]        
        predicted_label = [encoder.inverse_transform([val[1]])[0] for val in predicted_prob_class]
        
        log.info('[*] Completed the model prediction on test data')
        
        # Storing the model results in predicted results
        log.info('[*] Storing the model prediction results')

        X_test['predicted_datatype_l1'] = predicted_label
        X_test['predicted_probability_l1'] = predicted_prob
             
        # Subset the required columns based on the process type
        
        return X_test[['table_name', 'column_name', 'predicted_datatype_l1', 'predicted_probability_l1']]
    



import json
import logging
import re
import time

import pandas as pd

from langchain_core.prompts import PromptTemplate
from tqdm import tqdm

from intugle.core import settings
from intugle.core.llms.chat import ChatModelLLM
from intugle.core.utilities.llm_utils import generate_create_table_query, read_column_datatypes

from .prompts import (
    BUSINESS_GLOSSARY_PROMPTS,
    column_glossary,
    column_tag_glossary,
    table_glossary,
)
from .utils import (
    get_additional_context,
    manual_parsing,
    preprocess_profiling_df,
)

log = logging.getLogger(__name__)

tqdm.pandas()
line_count = settings.BG_CONFIG["THRESHOLD"]["LINE_COUNT"]
word_count = settings.BG_CONFIG["THRESHOLD"]["WORD_COUNT"]


TABLE_DESCRIPTION = settings.BG_CONFIG["GLOSSARY_GEN"]["TABLE_DESCRIPTION"]
COLUMN_GLOSSARY = settings.BG_CONFIG["GLOSSARY_GEN"]["COLUMN_GLOSSARY"]
COLUMN_TECHNICAL_GLOSSARY = settings.BG_CONFIG["GLOSSARY_GEN"]["COLUMN_TECHNICAL_GLOSSARY"]
BUSINESS_TAGS = settings.BG_CONFIG["GLOSSARY_GEN"]["BUSINESS_TAGS"]
USE_COLUMN_CONTEXT = settings.BG_CONFIG["GLOSSARY_GEN"]["USE_COLUMN_CONTEXT"]


def raw_json_parser(regex: str, data: str):
    match = re.search(regex, data, re.DOTALL)

    if match:
        inner_dict_str = match.group(1)
        inner_dict = json.loads(inner_dict_str)
        return inner_dict
    else:
        log.error(f"[!] Couldnot parse raw llm json : {data}")
        return ""


class BusinessGlossary:
    MAX_REGEX_RETRIES = 1

    TEMPLATE_NAME = "gpt-4o-mini"

    LLM_CONFIG_1 = {
        "temperature": 0.4,
        "stop": "END",
    }

    LLM_CONFIG_2 = {
        "temperature": 0.05,
        "stop": "END",
    }

    def __init__(self, profiling_data: pd.DataFrame, *args, **kwargs):

        profiling_data = preprocess_profiling_df(profiling_data.copy())
        self.profiling_data = profiling_data[
            [
                "table_name",
                "column_name",
                "sample_data",
                "datatype_l1",
            ]
        ]

        self.column_datatypes = (
            read_column_datatypes(
                dtype=self.profiling_data[
                    ["table_name", "column_name", "datatype_l1"]
                ].reset_index(drop=True)
            )
            if settings.BG_CONFIG["INCLUDE_DTYPES"]
            else {}
        )

        self.__table_glossary_llm = ChatModelLLM.build(
            model_name=settings.LLM_PROVIDER,
            llm_config=self.LLM_CONFIG_1,
            response_schemas=table_glossary,
            template_string=BUSINESS_GLOSSARY_PROMPTS[self.TEMPLATE_NAME][
                "TABLE_GLOSSARY_TEMPLATE"
            ],
            prompt_template=PromptTemplate,
        )

        self.__business_glossary_llm = ChatModelLLM.build(
            model_name=settings.LLM_PROVIDER,
            llm_config=self.LLM_CONFIG_2,
            response_schemas=column_glossary,
            template_string=BUSINESS_GLOSSARY_PROMPTS[self.TEMPLATE_NAME][
                "BUSINESS_GLOSSARY_TEMPLATE"
            ],
            prompt_template=PromptTemplate,
        )
        self.__business_tags_llm = ChatModelLLM.build(
            model_name=settings.LLM_PROVIDER,
            llm_config=self.LLM_CONFIG_2,
            response_schemas=column_tag_glossary,
            template_string=BUSINESS_GLOSSARY_PROMPTS[self.TEMPLATE_NAME][
                "BUSINESS_TAGS_TEMPLATE"
            ],
            prompt_template=PromptTemplate,
        )

        self.global_additional_context: str = kwargs.get(
            "global_additional_context", ""
        )
        self.table_additional_contexts: dict = kwargs.get(
            "table_additional_contexts", {}
        )

    @classmethod
    def fetch_table_glossary(
        cls,
        llm: ChatModelLLM,
        sql_statements: str,
        domain: str,
        table_name: str = "",
        additional_context: str = "",
    ):
        response, parsing_success, raw_response = llm.invoke(
            create_statements=sql_statements,
            domain=domain,
            table=table_name,
            additional_context=additional_context,
            metadata={
                "type": "table glossary",
                "table": table_name,
            },
        )

        if parsing_success:
            try:
                return response["table_glossary"], raw_response
            except Exception:
                response = raw_response

        result = manual_parsing(
            keys_name=["table_glossary", table_name],
            text=response,
        )
        return result, raw_response

    @classmethod
    def fetch_column_glossary(
        cls,
        llm: ChatModelLLM,
        sql_statements: str,
        domain: str,
        table_name: str,
        column_name: str,
        additional_context: str = "",
    ):
        # format_output = """
        # # Output the business glossary in the following structure:
        # ```json
        #     "column_glossary": { 
        #         "column_name_1": "A brief description of the purpose or role of column_name_1."},
        #         "column_name_2": "A brief description of the purpose or role of column_name_2."}
        #     }
        # ```        
        # """
        response, parsing_success, raw_response = llm.invoke(
            create_statements=sql_statements,
            domain=domain,
            column=column_name,
            additional_context=additional_context,
            metadata={
                "type": "column glossary",
                "column": f"{table_name}.{column_name}",
            },
        )

        if parsing_success:
            try:
                return response["column_glossary"], raw_response
            except Exception:
                response = raw_response

        result = manual_parsing(
            keys_name=["column_glossary", column_name], text=response
        )
        return result, raw_response

    @classmethod
    def fetch_business_tags(
        cls,
        llm: ChatModelLLM,
        sql_statements: str,
        domain: str,
        table_name: str,
        column_name: str,
        additional_context: str = "",
    ):
        # format_output = """
        #     # Output the tags in the following structure:
        #     ```json
        #     "column_tag_glossary":{
        #         "column_name": ["Tag1", "Tag2", "Tag3"],
        #         }
        #     ```
        # """
        response, parsing_success, raw_response = llm.invoke(
            create_statements=sql_statements,
            domain=domain,
            column=column_name,
            additional_context=additional_context,
            metadata={"type": "column tags", "column": f"{table_name}.{column_name}"},
        )

        if parsing_success:
            try:
                return response["column_tag_glossary"], raw_response
            except Exception:
                response = raw_response

        result = manual_parsing(
            keys_name=["column_tag_glossary", column_name], text=response
        )

        return result, raw_response

    def bg_tg(self, row, profiling_temp, domain, additional_context: str = ""):
        column = row["column_name"]
        table_name = row["table_name"]
        sql_statements_2 = generate_create_table_query(
            table_columns=[column],
            column_datatypes=self.column_datatypes,
            table_name=table_name,
            profiling_data=profiling_temp,
            columns_required=["sample_data"],
        )

        glossary, _ = self.fetch_column_glossary(
            llm=self.__business_glossary_llm,
            sql_statements=sql_statements_2,
            domain=domain,
            column_name=column,
            table_name=table_name,
            additional_context=additional_context,
        )
        tags, _ = self.fetch_business_tags(
            llm=self.__business_tags_llm,
            sql_statements=sql_statements_2,
            domain=domain,
            column_name=column,
            table_name=table_name,
            additional_context=additional_context,
        )

        # ()
        return pd.Series([glossary, tags], index=["business_glossary", "business_tags"])

    def __run_bg__(self, table_name: str, domain: str):
        st = time.time()

        # for table_name,profiling_temp in self.profiling_data.groupby(["table_name"]):
        sql_statements = generate_create_table_query(
            table_columns=self.profiling_data.column_name.tolist(),
            column_datatypes=self.column_datatypes,
            table_name=table_name,
            profiling_data=self.profiling_data,
            columns_required=["sample_data"],
        )

        additional_context_to_llm = get_additional_context(
            table_name=table_name,
            global_additional_context=self.global_additional_context,
            additional_table_context=self.table_additional_contexts.get(
                str(table_name), ""
            ),
        )

        table_glossary, _ = self.fetch_table_glossary(
            llm=self.__table_glossary_llm,
            sql_statements=sql_statements,
            domain=domain,
            table_name=table_name,
            additional_context=additional_context_to_llm,
        )
        # ()

        self.profiling_data.loc[
            self.profiling_data.index, ["business_glossary", "business_tags"]
        ] = self.profiling_data.progress_apply(
            self.bg_tg,
            axis=1,
            profiling_temp=self.profiling_data,
            domain=domain,
            additional_context=additional_context_to_llm,
        )
        self.profiling_data.loc[self.profiling_data.index, "table_glossary"] = table_glossary
        end = time.time()
        self.profiling_data.loc[self.profiling_data.index, "execution_time"] = (
            f"{end - st:.2f}"
        )
        return table_glossary, self.profiling_data

    def __call__(self, table_name: str, domain: str):
        table_glossary, glossary_df = self.__run_bg__(table_name, domain=domain)
        return table_glossary, glossary_df
